import socket
try:
    import socketserver
except:
    import SocketServer as socketserver
import logging
from openmath import openmath as om, convert as conv

from scscp.client import TimeoutError, CONNECTED
from scscp.server import SCSCPServer
from scscp.scscp import SCSCPQuit, SCSCPProtocolError
from scscp import scscp

# Supported functions
CD_SCSCP2 = ['get_service_description', 'get_allowed_heads', 'is_allowed_head']
CD_ARITH1 = {
    'abs'         : abs,
    'unary_minus' : lambda x  : -x,
    'minus'       : lambda x,y: x-y,
    'plus'        : lambda x,y: x+y,
    'divide'      : lambda x,y: x/y,
    'times'       : lambda x,y: x*y,
    'power'       : lambda x,y: x**y,
}

class SCSCPRequestHandler(socketserver.BaseRequestHandler):
    def setup(self):
        self.server.log.info("New connection from %s:%d" % self.client_address)
        self.log = self.server.log.getChild(self.client_address[0])
        self.scscp = SCSCPServer(self.request, logger=self.log)
        
    def handle(self):
        self.scscp.accept()
        while True:
            try:
                call = self.scscp.wait()
            except TimeoutError:
                continue
            except SCSCPQuit as e:
                self.log.info(e)
                break
            except ConnectionResetError:
                self.log.info('Client closed unexpectedly.')
                break
            except SCSCPProtocolError as e:
                self.log.info('SCSCP protocol error: %s.' % str(e))
                self.log.info('Closing connection.')
                self.scscp.quit()
                break
            self.handle_call(call)

    def handle_call(self, call):
        if (call.type != 'procedure_call'):
            raise SCSCPProtocolError('Bad message from client: %s.' % call.type, om=call.om())
        try:
            head = call.data.elem.name
            self.log.debug('Requested head: %s...' % head)
            
            if call.data.elem.cd == 'scscp2' and head in CD_SCSCP2:
                res = getattr(self, head)(call.data)
            elif call.data.elem.cd == 'arith1' and head in CD_ARITH1:
                args = [conv.to_python(a) for a in call.data.arguments]
                res = conv.to_openmath(CD_ARITH1[head](*args))
            else:
                self.log.debug('...head unknown.')
                return self.scscp.terminated(call.id, om.OMError(
                    om.OMSymbol('unhandled_symbol', cd='error'), [call.data.elem]))

            strlog = str(res)
            self.log.debug('...sending result: %s' % (strlog[:20] + (len(strlog) > 20 and '...')))
            return self.scscp.completed(call.id, res)
        except (AttributeError, IndexError, TypeError):
            self.log.debug('...client protocol error.')
            return self.scscp.terminated(call.id, om.OMError(
                om.OMSymbol('unexpected_symbol', cd='error'), [call.data]))
        except Exception as e:
            self.log.exception('Unhandled exception:')
            return self.scscp.terminated(call.id, 'system_specific',
                                             'Unhandled exception %s.' % str(e))

    def get_allowed_heads(self, data):
        return scscp.symbol_set([om.OMSymbol(head, cd='scscp2') for head in CD_SCSCP2]
                                    + [om.OMSymbol(head, cd='arith1') for head in CD_ARITH1],
                                    cdnames=['scscp1'])
    
    def is_allowed_head(self, data):
        head = data.arguments[0]
        return conv.to_openmath((head.cd == 'arith1' and head.name in CD_ARITH1)
                                    or (head.cd == 'scscp2' and head.name in CD_SCSCP2)
                                    or head.cd == 'scscp1')

    def get_service_description(self, data):
        return scscp.service_description(self.server.description)

class Server(socketserver.ThreadingMixIn, socketserver.TCPServer, object):
    allow_reuse_address = True
    
    def __init__(self, host='localhost', port=26133,
                     logger=None, description='Demo SCSCP server'):
        super(Server, self).__init__((host, port), SCSCPRequestHandler)
        self.log = logger or logging.getLogger(__name__)
        self.description = description
        
if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    logger = logging.getLogger('demo_server')
    srv = Server(logger=logger)
    try:
        srv.serve_forever()
    except KeyboardInterrupt:
        srv.shutdown()
        srv.server_close()




import socket
from openmath import convert, openmath as om
from .client import SCSCPClient
from . import scscp

def _conv_if_py(obj):
    if isinstance(obj, om.OMAny):
        return obj
    else:
        return convert.to_openmath(obj)

class SCSCPCLI(SCSCPClient):
    """ A synchronous CLI client for SCSCP """

    class Head(object):
        """ A callable remote procedure """
        def __init__(self, name, cd, cli):
            self._name = name
            self.cd = cd
            self._cli = cli
            self._om = om.OMSymbol(name, cd=cd)
        def __call__(self, data, cookie=False, **opts):
            res = self._cli._call_wait(om.OMApplication(self._om, map(_conv_if_py, data)),
                                           cookie, **opts)
            if res.type == 'procedure_completed':
                try:
                    return convert.to_python(res.data)
                except ValueError:
                    return res.data
            elif res.type == 'procedure_terminated':
                raise scscp.SCSCPProtocolError('Server returned error: %s.' % res.data.name.name,
                                                   res.data)
            else:
                raise scscp.SCSCPProtocolError('Unexpected response.', resp.om())
    
    class CD(object):
        """ A content dictionary, implemented as a namespace """
        def __init__(self, name, cli):
            self._name = name
            self._cli = cli
        def __getattr__(self, attr):
            return SCSCPCLI.Head(attr, self._name, self._cli)
        def _get_head(self, name):
            head = self.__dict__.get(name)
            if head is None:
                head = self.__dict__[name] = getattr(self, name)
            return head
        def __repr__(self):
            return repr([ k for k in self.__dict__ if k[0] != '_' ])
        def __contains__(self, head):
            return head in self.__dict__
    
    class Heads(object):
        """ A list of remote procedures """
        def __init__(self, cli):
            self._cli = cli
        def __getattr__(self, attr):
            return SCSCPCLI.CD(attr, self._cli)
        def _get_cd(self, name):
            cd = self.__dict__.get(name)
            if cd is None:
                cd = self.__dict__[name] = getattr(self, name)
            return cd
        def __repr__(self):
            return repr({ k:v for k,v in self.__dict__.items() if k[0] != '_' })
        def __contains__(self, cd):
            return cd in self.__dict__
            
    
    def __init__(self, host, port=26133, populate=True):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        super(SCSCPCLI, self).__init__(s)
        self.heads = self.Heads(self)
        self.connect()
        if populate:
            self.populate_heads()

    def _call_wait(self, data, cookie=False, **opts):
        call = self.call(data, cookie, **opts)
        resp = self.wait()
        
        if resp.id != call.id:
            raise scscp.SCSCPProtocolError("Wrong call id (expected %s, got %s)."
                                               % (call.id, resp.id), resp.om())

        return resp

    def populate_heads(self):
        heads = self._call_wait(scscp.get_allowed_heads())
        if heads.type == 'procedure_terminated':
            raise scscp.SCSCPProtocolError("Failed to get heads (%s)." % heads.data.name.name,
                                               heads.data)
        elif heads.type != 'procedure_completed':
            raise scscp.SCSCPProtocolError("Server gave unexpected response.", heads.om())
    
        try:
            for symbol in heads.data.arguments:
                if isinstance(symbol, om.OMSymbol):
                    self.heads._get_cd(symbol.cd)._get_head(symbol.name)
                elif symbol.elem.name == 'CDName':
                    self.heads._get_cd(symbol.arguments[0].string)
                else:
                    continue
        except (AttributeError, IndexError):
            raise scscp.SCSCPProtocolError("Server gave unexpected response.", heads.data)

    def is_allowed_head(self, name, cd):
        return self.heads.scscp2.is_allowed_head([om.OMSymbol(name, cd)])
    
    def get_description(self):
        return '\n'.join(a.string for a in self.heads.scscp2.get_service_description([]).arguments)



import re
from .scscp import SCSCPConnectionError

class ProcessingInstruction():
    PI_regex = re.compile(b"<\?scscp\s+(.{0,4084}?)\?>", re.S)
    PI_regex_full = re.compile(b'^<\?scscp(?:\s+(?P<key>\w+))?(?P<attrs>(?:\s+\w+=".*?")*)\s*\?>$')
    PI_regex_attr = re.compile(b'(\w+)="(.*?)"')

    @classmethod
    def parse(cls, bytes):
        match = cls.PI_regex_full.match(bytes)
        if match:
            key = (match.group('key') or b'').decode('ascii')
            attrs = { k.decode('ascii'): v
                          for k, v in cls.PI_regex_attr.findall(match.group('attrs')) }
            return cls(key, **attrs)
        else:
            raise SCSCPConnectionError("Bad SCSCP processing instruction %s." % bytes)

    def __init__(self, key='', **attrs):
        self.key = key
        self.attrs = attrs

    def __bytes__(self):
        return b'<?scscp %s %s ?>' % (self.key.encode(),
                                         b' '.join(b'%s="%s"' % (k.encode(), v)
                                                      for k,v in self.attrs.items()))

    def __str__(self):
        return '<?scscp %s %s ?>' % (self.key,
                                         ' '.join('%s="%s"' % (k, v.decode())
                                                      for k,v in self.attrs.items()))
    
    def __repr__(self):
        return 'ProcessingInstruction: %s' % self



from .client import SCSCPPeer, SCSCPPeerOM, _assert_status, INITIALIZED, CONNECTED
from .scscp import SCSCPConnectionError, SCSCPProcedureMessage

class SCSCPServerBase(SCSCPPeer):
    """
    A simple SCSCP synchronous server, with no understanding of OpenMath.
    """
    
    def __init__(self, socket, timeout=30, logger=None):
        super(SCSCPServerBase, self).__init__(socket, timeout, logger, me="Server", you="Client")

    @_assert_status(INITIALIZED, "Session already opened.")
    def accept(self, timeout=None):
        """ SCSCP handshake """
        self._send_PI(scscp_versions=b'1.3')
        
        pi = self._get_next_PI([''], timeout=timeout)
        if pi.attrs.get('version') != b'1.3':
            self.quit()
            raise SCSCPConnectionError("Client sent unexpected response.", pi)
        
        self._send_PI(version=b'1.3')

        self.status = CONNECTED

        
class SCSCPServer(SCSCPServerBase, SCSCPPeerOM):
    """
    A simple SCSCP synchronous server.
    """
    def wait(self, timeout=-1):
        return SCSCPProcedureMessage.from_om(self.receive(timeout))

    def completed(self, id, data, **info):
        comp = SCSCPProcedureMessage.completed(id, data, **info)
        self.send(comp.om())
        return comp

    def terminated(self, id, error, msg=None, **info):
        term = SCSCPProcedureMessage.terminated(id, error, msg, **info)
        self.send(term.om())
        return term



import uuid
import openmath.openmath as om

class SCSCPError(RuntimeError):
    pass
class SCSCPConnectionError(SCSCPError):
    def __init__(self, msg, pi=None):
        super(SCSCPConnectionError, self).__init__(msg)
        self.pi = pi
class SCSCPCancel(SCSCPError):
    pass
class SCSCPQuit(SCSCPError):
    def __init__(self, msg, reason=''):
        super(SCSCPQuit, self).__init__(msg)
        self.reason = reason
class SCSCPProtocolError(SCSCPError):
    def __init__(self, msg, om=None):
        super(SCSCPProtocolError, self).__init__(msg)
        self.om = om

### SCSCP1 content dictionary

class SCSCPProcedureMessage(object):
    options = {
        'debuglevel'     : om.OMInteger,
        'max_memory'     : om.OMInteger,
        'min_memory'     : om.OMInteger,
        'return_cookie'  : om.OMString,
        'return_object'  : om.OMString,
        'return_nothing' : om.OMString,
        'runtime'        : om.OMInteger,
    }

    infos = {
        'memory'  : om.OMInteger,
        'runtime' : om.OMInteger,
        'message' : om.OMString,
    }

    # Whether error message is mandatory
    errors = {
        'memory'          : False,
        'runtime'         : False,
        'system_specific' : True,
    }

    def __init__(self, type, data, id=None, params=None):
        self.type = type
        self.id = str(id or uuid.uuid1())
        self.params = params or []
        self.data = data

    @classmethod
    def call(cls, data, id=None, **opts):
        opts = [(om.OMSymbol('option_' + k, cd='scscp1'),
                    cls.options[k](v))
                    for k, v in opts.items() if v is not None and v is not False]
        return cls('procedure_call', data, id, opts)

    @classmethod
    def _w_info(cls, type, id, data, **info):
        info = [(om.OMSymbol('info_' + k, cd='scscp1'),
                    cls.infos[k](v))
                    for k, v in info.items() if v is not None]
        return cls(type, data, id, info)

    @classmethod
    def completed(cls, id, data, **info):
        return cls._w_info('procedure_completed', id, data, **info)

    @classmethod
    def terminated(cls, id, error, msg=None, **info):
        if not isinstance(error, om.OMError):
            if cls.errors[error] and msg is None:
                raise RuntimeError('Must give an error message')
            error = om.OMError(om.OMSymbol('error_' + error, cd='scscp1'), [om.OMString(msg)])
        return cls._w_info('procedure_terminated', id, error, **info)

    def __repr__(self):
        return "SCSCPProcedureMessage %s#%s" % (self.type, self.id)

    def __eq__(self, other):
        return (isinstance(other, SCSCPProcedureMessage)
                    and self.type == other.type and self.id == other.id
                    and self.params == other.params and self.data == other.data)
    
    def om(self):
        return om.OMObject(
            om.OMAttribution(
                om.OMAttributionPairs([(om.OMSymbol('call_id', cd='scscp1'), om.OMString(self.id))] + self.params),
                om.OMApplication(
                    om.OMSymbol(self.type, cd='scscp1'),
                    [self.data]
                )
            )
        )
        
    @classmethod
    def from_om(cls, obj):
        if not (isinstance(obj, om.OMObject)
                and isinstance(obj.omel, om.OMAttribution)):
            raise SCSCPProtocolError('Bad SCSCP procedure message.', obj)
        params = obj.omel.pairs.pairs

        try:
            index, id = next((i,p) for i,p in enumerate(params)
                                 if p[0].name == 'call_id' and p[0].cd == 'scscp1')
        except StopIteration:
            raise SCSCPProtocolError('SCSCP procedure message does not contain id.', obj)
        if not isinstance(id[1], om.OMString):
            raise SCSCPProtocolError('Bad SCSCP procedure message.', obj)
        params.pop(index)
        id = id[1].string

        if not (isinstance(obj.omel.obj, om.OMApplication)
                    and obj.omel.obj.elem.cd == 'scscp1'
                    and len(obj.omel.obj.arguments) == 1):
            raise SCSCPProtocolError('Bad SCSCP procedure message.', obj)
        type = obj.omel.obj.elem.name
        data = obj.omel.obj.arguments[0]

        return cls(type, data, id, params)
        

### SCSCP2 content dictionary

def _apply(cmd, data):
    return om.OMApplication(om.OMSymbol(cmd, cd='scscp2'), data)

def symbol_set(symbols=[], cdnames=['scscp1', 'scscp2'], cdurls=[], groupnames=[], groupurls=[]):
    aggregations = (
        (cdnames, om.OMSymbol('CDName', cd='meta')),
        (cdurls, om.OMSymbol('CDURL', cd='meta')),
        (groupnames, om.OMSymbol('CDGroupName', cd='metagrp')),
        (groupurls, om.OMSymbol('CDGroupURL', cd='metagrp')),
    )
    return _apply('symbol_set', symbols
                      + sum([[om.OMApplication(symbol, [om.OMString(x)]) for x in list]
                                 for (list, symbol) in aggregations], []))

def get_allowed_heads():
    return _apply('get_allowed_heads', [])

def is_allowed_head(name, cd):
    return  _apply('is_allowed_head', [om.OMSymbol(name, cd)])

def store(data, peristent=False):
    return _apply('store_' + ('persistent' if persistent else 'session'), [data])

def retrieve(url):
    return _apply('retrieve', [om.OMReference(url)])

def unbind(url):
    return _apply('unbind', [om.OMReference(url)])

def get_transient_cd(name):
    return _apply('get_transient_cd', [om.OMApply(
        om.OMSymbol('CDName', cd='meta'), [om.OMString(name)]
    )])

def get_signature(name, cd):
    return  _apply('get_signature', [om.OMSymbol(name, cd)])

def signature(name, cd, symbol_sets=None, min=None, max=None):
    if isinstance(symbol_sets, list):
        n = len(symbol_sets)
        if min is None:
            min = om.OMInteger(n)
        else:
            assert min <= n, "Too few parameters"
            min = om.OMInteger(min)
        assert max is None or max == n, "Too many parameters"
        max = om.OMInteger(n)
        symbol_sets = om.OMApplication(om.OMSymbol('list', cd='list1'), symbol_sets)
    else:
        min = om.OMInteger(0 if min is None else min)
        max = om.OMSymbol('infinity', cd='nums1') if max is None else om.OMInteger(max)
        if symbol_sets is None:
            symbol_sets = om.OMSymbol('symbol_set_all', cd='scscp2')
    return _apply('signature', [om.OMSymbol(name, cd), min, max, symbol_sets])

def get_service_description():
    return _apply('get_service_description', [])

def service_description(*desc):
    return _apply('service_description', [om.OMString(d) for d in desc])

def no_such_transient_cd(cd):
    return om.OMError(om.OMSymbol('no_such_transient_cd', cd='scscp2'), [om.OMString(cd)])



import logging
from pexpect import fdpexpect, TIMEOUT, EOF
from openmath import encoder, decoder
from . import scscp
from .scscp import SCSCPConnectionError, SCSCPQuit, SCSCPCancel, SCSCPProcedureMessage
from .processing_instruction import ProcessingInstruction as PI

class TimeoutError(RuntimeError):
    """ Client/Server timeout """
    pass
    
INITIALIZED=0
CONNECTED=1
CLOSED=2

def _assert_status(status, msg=None):
    def wrap(fun):
        def wrapper(self, *args, **kwds):
            if self.status != status:
                raise RuntimeError(msg or "Bad status %d." % self.status)
            return fun(self, *args, **kwds)
        return wrapper
    return wrap
_assert_connected = _assert_status(CONNECTED, "Not connected.")


class SCSCPPeer(object):
    """
    Base class for SCSCP client and server
    """
    
    def __init__(self, socket, timeout=30, logger=None, me='Client', you='Server'):
        self.socket = socket
        self.stream = fdpexpect.fdspawn(socket, timeout=timeout)
        self.status = INITIALIZED
        self.log = logger or logging.getLogger(__name__)
        self.me, self.you = me, you

    def _get_next_PI(self, expect=None, timeout=-1):
        while True:
            try:
                self.stream.expect(PI.PI_regex, timeout=timeout)
            except TIMEOUT:
                raise TimeoutError("%s took too long to respond." % self.you)
            except EOF:
                raise ConnectionResetError("%s closed unexpectedly." % self.you)

            try:
                pi = PI.parse(self.stream.after)
            except SCSCPConnectionError:
                self.quit()
                raise
            self.log.debug("Received PI: %s" % pi)

            if expect is not None and pi.key not in expect:
                if pi.key == 'quit':
                    self.quit()
                    reason = pi.attrs.get('reason')
                    raise SCSCPQuit("%s closed session (reason: %s)." % (self.you, reason), reason)
                if pi.key == '' and 'info' in pi.attrs:
                    self.log.info("SCSCP info: %s" % pi.attrs.get('info').decode())
                    continue
                else:
                    raise SCSCPConnectionError("%s sent unexpected message: %s" % (self.you, pi.key), pi)
            else:
                return pi


    def _send_PI(self, key='', **kwds):
        pi = PI(key, **kwds)
        self.log.debug("Sending PI: %s" % pi)
        return self.socket.send(bytes(pi))

    @_assert_connected
    def send(self, msg):
        """ Send SCSCP message """
        self._send_PI('start')
        try:
            self.socket.send(msg)
        except:
            self._send_PI('cancel')
            raise
        else:
            self._send_PI('end')

    @_assert_connected
    def receive(self, timeout=-1):
        """ Receive SCSCP message """
        msg = b""
        pi = self._get_next_PI(['start'], timeout=timeout)
        while True:
            pi = self._get_next_PI(['end', 'cancel'], timeout=timeout)
            if pi.key == 'cancel':
                raise SCSCPCancel('%s canceled transmission' % self.you)
            
            msg += self.stream.before
            return msg

    @_assert_connected
    def quit(self, reason=None):
        """ Send SCSCP quit message """
        kwds = {} if reason is None else { 'reason': None }
        try:
            self._send_PI('quit', **kwds)
            self.socket.close()
        except ConnectionError:
            pass
        finally:
            self.status = CLOSED

    @_assert_connected
    def info(self, info):
        """ Send SCSCP info message """
        self._send_PI(info=info)


class SCSCPPeerOM(SCSCPPeer):
    """
    Base class for SCSCP client and server understanding OpenMath
    """
    
    def receive(self, timeout=-1):
        msg = super(SCSCPPeerOM, self).receive(timeout)
        return decoder.decode_bytes(msg)
        
    def send(self, om):
        return super(SCSCPPeerOM, self).send(encoder.encode_bytes(om))

    
class SCSCPClientBase(SCSCPPeer):
    """
    A simple SCSCP synchronous client, with no understanding of OpenMath.
    """
    
    def __init__(self, socket, timeout=30, logger=None):
        super(SCSCPClientBase, self).__init__(socket, timeout, logger, me="Client", you="Server")

    @_assert_status(INITIALIZED, "Session already opened.")
    def connect(self, timeout=None):
        """ SCSCP handshake """
        
        pi = self._get_next_PI([''], timeout=timeout)
        if ('scscp_versions' not in pi.attrs
                or b'1.3' not in pi.attrs['scscp_versions'].split()):
            self.quit()
            raise SCSCPConnectionError("Unsupported SCSCP versions %s." % pi.attrs.get('scscp_versions'), pi)
        
        self.service_info = pi.attrs

        self._send_PI(version=b'1.3')

        pi = self._get_next_PI([''], timeout=timeout)
        if pi.attrs.get('version') != b'1.3':
            self.quit()
            raise SCSCPConnectionError("Server sent unexpected response.", pi)

        self.status = CONNECTED

    @_assert_connected
    def terminate(self, id):
        """ Send SCSCP terminate message """
        self._send_PI('terminate', call_id=id)


class SCSCPClient(SCSCPClientBase, SCSCPPeerOM):
    """
    A simple SCSCP synchronous client.
    """
    def wait(self, timeout=-1):
        return SCSCPProcedureMessage.from_om(self.receive(timeout))

    def call(self, data, cookie=False, **opts):
        if cookie:
            opts['return_cookie'] = True
        elif cookie is None:
            opts['return_nothing'] = True
        else:
            opts['return_object'] = True
        call = SCSCPProcedureMessage.call(data, id=None, **opts)
        self.send(call.om())
        return call



from setuptools import setup, find_packages

setup(
    name='scscp',
    version='0.1.0a1',
    description='Implementation of the SCSCP protocol',
    url='https://github.com/OpenMath/py-scscp',
    author='Luca De Feo',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    keywords='openmath scscp',
    packages=find_packages(),
    install_requires=['openmath', 'pexpect'],
)
